#include "GRCPrinter.hpp"
#include <cassert>

namespace AST {

  void GRCDot(std::ostream &o, GRCNode *n, Module *m, Nodemap &refmap)
  {
    int mxnode;

    assert(n);
    assert(m);
    assert(m->symbol);

    mxnode=-1;
    for (Nodemap::iterator i=refmap.begin(); i!= refmap.end(); i++)
      if(mxnode < i->second) mxnode = i->second;
    
    GRCDP p(o, refmap, mxnode+1);
    o << "digraph " << m->symbol->name << " {" << std::endl;
    o << "size=\"7.5,10\"\n";
    p.visit(n);
    p.drawparents();
    o << "}" << std::endl;
  }


  void GRCDP::drawparents(){

    int sz, i;
    GRCNode *n;

    if(!drawpar) return;

    sz=topo.size();
    for(i=0;i<sz;i++){
      n=topo[i];
      for (vector<GRCNode*>::iterator j = n->predecessors.begin();
	   j != n->predecessors.end(); j++) {
	if (*j && refmap.count(*j) > 0) {
	  o << 'n' << refmap[n] << " -> n" << refmap[*j];	
	  //	  o << "[label=\"" << j - n->predecessors.begin() << "\" color=green constraint=false]";
	  o<<"[color=green constraint=false]\n";
	  o << '\n';
	} else {
	  o<< 'n'<< refmap[n] << " -> n" << nextnum;
	  //o<<"[label=\""<<j-n->predecessors.begin()<<"\" color=green ]"<<'\n';
	  o<<"[color=green]\n";
	  o<< 'n'<<nextnum++<<" [label=\"o\"]"<<'\n';
	}
      }
    }
  }

  void GRCDP::visit(GRCNode *n) {
    assert(n);

    if(reached.count(n)>0) return;
    reached[n]=1;

    if ( refmap.count(n) == 0 ) {
      refmap[n] = nextnum++;
    }
    
    // Visit each of our successors first
    for ( vector<GRCNode*>::iterator j = n->successors.begin() ;
	  j != n->successors.end() ; j++ ) {
      if (*j) {
	visit(*j);
	o << 'n' << refmap[n] << " -> n" << refmap[*j];
	if ( n->successors.size() > 1)
	  o << "[label=\"" << j - n->successors.begin() << "\"]";
	o << '\n';
      } else {
	o<< 'n'<< refmap[n] << " -> n" << nextnum<<"[label=\""<<j-n->successors.begin()<<"\"]"<<'\n';
	o<< 'n'<<nextnum++<<" [label=\"o\"]"<<'\n';
      }
    }
    
    topo.push_back(n);

    // Print a definition for the node
    
    mynum = refmap[n];
    o << 'n' << mynum << ' ';
    n->welcome(*this);
    
  }

  void GRCDP::drawSTlink(GRCNode *g, STNode *s)
  {

    if(!drawblue) return;

    if(refmap.find(s)==refmap.end()) refmap[s] = nextnum++;
    o << 'n' << refmap[g] << " -> n" << refmap[s];
    o << "[color=blue constraint=false]";// style=dashed]";
    o << '\n';
  }

  Status GRCDP::visit(Switch &s) {
    o << "[label=\"" << mynum << " switch "<<refmap[s.st]<<" \" shape=triangle]\n";
    drawSTlink(&s,s.st);
    return Status();
  }


  Status GRCDP::visit(Test &s) {
    o << "[label=\"" << mynum << " test";
    s.predicate->welcome(ep);
    o << "\" shape=diamond]\n";
    return Status();
  }

  Status GRCDP::visit(Terminate &s) {
    o << "[label=\"" << mynum << " terminate " << s.code << "\" shape=octagon]\n";
    return Status();
  }

  Status GRCDP::visit(Sync &s) {
    o << "[label=\"" << mynum << " sync";
    if(s.st) o<<" "<<refmap[s.st];
    o<<"\" shape=box]\n";
    return Status();
  }

  Status GRCDP::visit(Fork &s) {
    o << "[label=\"" << mynum << " fork\" shape=triangle]\n";
    return Status();
  }

  Status GRCDP::visit(Action &s) {
    o << "[label=\"" << mynum << " action ";
    s.body->welcome(ep);
    o << "\" shape=box]\n";
    return Status();
  }

  Status GRCDP::visit(Enter &s) {
    o << "[label=\"" << mynum << " enter "<<refmap[s.st];
    o << "\" shape=house]\n";
    drawSTlink(&s, s.st);
    return Status();
  }

  Status GRCDP::visit(Leave &s) {
    o << "[label=\"" << mynum << " leave "<<refmap[s.st];
    o << "\" shape=invhouse]\n";
    drawSTlink(&s, s.st);
    return Status();
  }

  Status GRCDP::visit(STexcl &s) {
    o << "[label=\"" << mynum << " #";
    o << "\" shape=egg]\n";
    return Status();
  }

  Status GRCDP::visit(STref &s) {
    o << "[label=\"" << mynum << " ";
    if(s.isabort()) o << "A";
    if(s.issuspend()) o << "S";
    o << "\" shape=egg]\n";
    return Status();
  }

  Status GRCDP::visit(STpar &s) {
    o << "[label=\"" << mynum << " ||";
    o << "\" shape=egg]\n";
    return Status();
  }
  
  Status GRCDP::visit(STleaf &s) {
    o << "[label=\"" << mynum << " ";
    if(s.isfinal()) o<<"*";
    o << "\" shape=box]\n";
    return Status();
  }

  Status GRCDP::visit(TopGRC &s) {
     o << "[label=\"" << mynum << " TOP\" shape=triangle]\n";
     return Status();
  }


  Status GRCDP::visit(Nop &s){
    o << "[label=\"" << mynum << " ";
    if(s.isflowin()) o<<"*"; else 
      if(s.isshortcut()) o<<"#"; else o<<"\\n"<<s.code;
    o << "\" shape=circle]\n";
    return Status();
  }
  Status GRCDP::visit(RecT1&s) {
    o << "[label=\"T1 " << mynum;
    o << "\" shape=box]\n";
    return Status();
  }


  Status GRCDP::visit(DefineSignal &s){
    o<< "[label=\"DefS "<<mynum<<"\\n";
    for(SymbolTable::iterator isym = s.signals->begin();
	isym != s.signals->end(); isym++)
      o<<(*isym)->name<<" ";

    o<<"\"]\n";
    return Status();
  }
  Status GRCDP::visit(UndefineSignal &s){
    o<< "[label=\"UndS\\n";
    for(SymbolTable::iterator isym = s.signals->begin();
	isym != s.signals->end(); isym++)
      o<<(*isym)->name<<" ";
    o<<"\"]\n";
    return Status();
  }

}
