#include "IR.hpp"
#include "AST.hpp"
#include "GRCPrinter.hpp"
#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <set>

namespace AST {

  class GRC2trillian: public Visitor {
    std::ostream &o;

    int debug;
    int useold;

    TopGRC *top;
    SymbolTable *sigs;
    STNode *stroot;
    STleaf  *boot, *finished;
    Switch *top_sw;

    Nodemap &refmap;
    map<STNode*, int> states;
    vector<GRCNode*> topo;
    map<GRCNode*,int> sched;
    std::set<GRCNode*> dfs_black;
    map<SignalSymbol*,int> signals;
    map<SignalSymbol*,std::set<GRCNode*> > receivers;

  public:
    GRC2trillian(std::ostream &oo, TopGRC *top, SymbolTable *sigs, Nodemap &refmap): o(oo), top(top), sigs(sigs), refmap(refmap) {}
    virtual ~GRC2trillian() {}
    
    void visit (GRCNode *);

    Status visit(Switch &);
    Status visit(Test &);
    Status visit(Terminate &);
    Status visit(Sync &);
    Status visit(Fork &);
    Status visit(Action &);
    Status visit(Enter &);
    Status visit(Leave &);

    Status visit(TopGRC &){return Status();}
    Status visit(STexcl &){return Status();}
    Status visit(STref &){return Status();}
    Status visit(STpar &){return Status();}
    Status visit(STleaf &){return Status();}

    
    void collect_rcvs(GRCNode *n);
    void dfs(GRCNode *n);
    void init();

    void sched_children(GRCNode *);
    void dotick();
    void execute();

    void doenter(STNode *n);
    void doleave(STNode *n);
    int doswitch(STNode *n);

  };
  
  void GRC2trillian::visit (GRCNode *n)
  {
  }

  void GRC2trillian::doenter(STNode *n)
  {

    STNode *par;
    STexcl *ex;
    int i, ch, sz;

    if(n==stroot) return;

    par=(STNode*)n->predecessors[0];
    
    if((ex=dynamic_cast<STexcl*>(par))==NULL) return; // not good

    // look to parent; which child am I ?
    ch=-1;
    sz=par->successors.size();
    for(i=0;i<sz;i++)
      if(n==par->successors[i]) {ch=i; break;}

    states[par]=i;

  }
  void GRC2trillian::doleave(STNode *n)
  { 
  }

  int  GRC2trillian::doswitch(STNode *n) // returns the # of active child
  {
    if ((dynamic_cast<STexcl*>(n))==NULL)
      std::cerr<<"warning: testing not excl node\n";
    return states[n];
  }

  void GRC2trillian::collect_rcvs(GRCNode *n)
  {
    Test *tst;
    LoadSignalExpression *se;

    if(!n) return;
    if((tst=dynamic_cast<Test*>(n))){
      if((se=dynamic_cast<LoadSignalExpression*>(tst->predicate))){
	receivers[se->signal].insert(n);
      }
    }

    for(vector<GRCNode*>::iterator i=n->successors.begin(); i!=n->successors.end();i++){
      collect_rcvs(*i);
    }
  }

  void GRC2trillian::dfs(GRCNode *n){
    int i;
    
    Action *ac;
    Emit *e;
    std::set<GRCNode*>::iterator r;

    if(!n) return;
    if(dfs_black.count(n)>0) return;
    dfs_black.insert(n); // color black
    for(i=0;i<(int)n->successors.size();i++)
      dfs(n->successors[i]);

    ac = dynamic_cast<Action*>(n);
    if (ac)
      {
	e = dynamic_cast<Emit*>(ac->body);
	if (e){  //add arc
	  for(r = receivers[e->signal].begin(); r != receivers[e->signal].end(); r++)
	    dfs(*r);
	}
      }


    topo.push_back(n);
  }

  void GRC2trillian::init()
  {
   
    stroot=dynamic_cast<STNode*>(top->successors[0]);
    assert(stroot);
    boot=dynamic_cast<STleaf*>(stroot->successors[0]);
    assert(boot);
    finished=dynamic_cast<STleaf*>(stroot->successors[2]);
    assert(finished);

    doenter(boot);

    top_sw=dynamic_cast<Switch*>(top->successors[1]);
    assert(top_sw);
    std::cerr<<"will collect rcvrs\n";
    collect_rcvs(top_sw);
    std::cerr<<"will dfs\n";
    dfs(top_sw);
    std::cerr<<"init ok\n";

    if(0)for (vector<GRCNode*>::iterator i = topo.begin(); i != topo.end(); i++)
      std::cerr<<"Node "<<refmap[*i]<<'\n';

  }

   void GRC2trillian::dotick()
   {
     int i, tsz;

     SignalSymbol *ss;

     // clear all signals
     for(SymbolTable::iterator isym = sigs->begin(); isym != sigs->end(); isym++){
       if((ss=dynamic_cast<SignalSymbol*>(*isym))){
	 //	std::cerr<<"found signal"<<ss->name<<'\n';
	signals[ss]=0;
       }
     }

     tsz=topo.size();

     for(i=0;i<tsz;i++)sched[topo[i]]=0;
     sched[topo[tsz-1]]=1;

     for(i=tsz-1; i>=0; i--){
       if(!(sched[topo[i]])) continue;
       topo[i]->welcome(*this);
     }

   }

  void GRC2trillian::execute()
  {
    int ntick;

    debug=0;
    useold=0;

    init();
    ntick=0;
    do {
      std::cerr<<"dotick "<<ntick++<<" ... ... ... ...\n";
      dotick();
    } while (( states[stroot]<2) && (ntick <=10));
  }

  void GRC2trillian::sched_children(GRCNode *n){
    int i, sz;
    sz=n->successors.size();
    for (i=0;i<sz;i++)
      sched[n->successors[i]]=1;
  }


  // SWITCH
  
  Status GRC2trillian::visit(Switch &s)
  {
    int act_ch;

    if(debug)std::cerr<<"switch "<<refmap[&s]<<'\n';
    act_ch = doswitch(s.st);
    assert(act_ch!=-1);
    if(debug)      std::cerr<<"  active child:"<<act_ch<<'\n';
    sched[s.successors[act_ch]]=1;
    return Status();
  }

  // ENTER and LEAVE

  Status GRC2trillian::visit(Enter &s)
  {
    doenter(s.st);
    sched_children(&s);
    if(debug)    std::cerr<<"enter "<<refmap[&s]<<'\n';
    return Status();
  }

  Status GRC2trillian::visit(Leave &s)
  {
    doleave(s.st);
    sched_children(&s);
    if(debug)    std::cerr<<"leave "<<refmap[&s]<<'\n';
    return Status();
  }

  // ACTION (emit signal) and TEST (check signal)

  Status GRC2trillian::visit(Action &s)
  {

    Emit *e; SignalSymbol *ss;

    sched_children(&s);

    e=dynamic_cast<Emit*>(s.body);
    if(e){
      ss=e->signal;
      signals[ss]=1;
      std::cerr<<"emit "<<refmap[&s]<<" "<<ss->name<<'\n';
    } else std::cerr << "unknown action" <<'\n';
    return Status();
    
  }

  Status GRC2trillian::visit(Test &s)
  {
    
    LoadSignalExpression *sexpr; SignalSymbol *ss;
    int sigpres;

    sexpr=dynamic_cast<LoadSignalExpression*>(s.predicate);
    if(sexpr){
      ss=sexpr->signal;
      sigpres=signals[ss]; // 0 or 1

      assert(s.successors.size()==2);
      sched[s.successors[sigpres]]=1; // child 1 is the Y branch ?
      std::cerr<<"test "<<refmap[&s]<<" "<<ss->name<<":"<<sigpres<<'\n';
    }


    return Status(); 
  }

  // TERMINATE, SYNC and FORK

  Status GRC2trillian::visit(Terminate &s)
  {
    sched_children(&s);
    if(debug)    std::cerr<<"term "<<refmap[&s]<<'\n';
    return Status();
  }

  Status GRC2trillian::visit(Sync &s)
  {
    int i,sz,mxcode;
    Terminate * t;
    GRCNode * child;

    mxcode = 0;
    sz = s.predecessors.size();
    
    for (i = 0; i < sz; i++){
      assert(t = dynamic_cast<Terminate*>(s.predecessors[i]));
      if ((sched[t]) && (t->code>mxcode))
	mxcode = t->code;
    }

    sz = s.successors.size();
    assert(mxcode<sz);

    child=s.successors[mxcode];
    sched[child] = 1;

    if(debug)    std::cerr<<"sync "<<refmap[&s]<<" code:"<<mxcode<<'\n';

    return Status();

  }

  Status GRC2trillian::visit(Fork &s)
  {
    sched_children(&s);
    if(debug)    std::cerr<<"fork "<<refmap[&s]<<'\n';
    return Status();
  }

}

int main(int argc, char *argv[])
{

  if(argc<3) {std::cerr<<"Usage: ... grc_file dot_file\n"; return -1;}

  std::ifstream grcfile(argv[1]); if(!grcfile.good()) return -1;
  std::ofstream dotfile(argv[2]); if(!dotfile.good()) return -1;

  try {

    IR::XMListream r(grcfile);
    IR::Node *n;
    r >> n;
    grcfile.close();

    AST::Modules *mods = dynamic_cast<AST::Modules*>(n);
    if (!mods) throw IR::Error("Root node is not a Modules object");

    for ( std::vector<AST::Module*>::iterator i = mods->modules.begin() ;
	  i != mods->modules.end() ; i++ ) {
      assert(*i);

      AST::TopGRC *gn = dynamic_cast<AST::TopGRC*>((*i)->body);
      assert(gn);
      AST::SymbolTable *sig = dynamic_cast<AST::SymbolTable*>((*i)->signals);
      assert(sig);

      std::cout << "File read - OK"<<std::endl;

      AST::Nodemap refmap;
      AST::GRCDot(dotfile,(AST::GRCNode *)gn,(*i),refmap);
      dotfile.close();


      AST::GRC2trillian tw (std::cout, gn, sig, refmap);

      tw.execute();
    }

  } catch (IR::Error &e) {
    std::cerr << e.s << std::endl;
    exit(-1);
  }

  return 0;

}
