#include "grc2zaphod.hpp"
#include <stdio.h>

using namespace AST;

namespace ZAPHOD {

  // builds the topological list
  // for each node it remembers the corresponding "Sync"

  stack<Sync *> sync_stack;
  
  void GRC2zaphod::dfs(GRCNode *n){
    int i;
    Sync *sy;
    Fork *fk;
    Switch *sw;


    // vector<GRCNode*>::iterator ip;    
    
    if(!n) return;
    
    if (dynamic_cast<STNode *>(n)) return;

    if(dfs_black.count(n)>0) return;
    dfs_black.insert(n); // color black
    for(i=0;i<(int)n->successors.size();i++)
      dfs(n->successors[i]);

    // now it returns
    topo.push_back(n);
  
    if(sync_stack.empty()){
      next_sync[n] = NULL;
      // std::cerr << "Node "<<refmap[n]<<" has no sync\n";
    }else {
      next_sync[n]=sync_stack.top();
      // std::cerr<<"Node "<<refmap[n]<<" goes to sync "<<refmap[next_sync[n]]<<'\n';
    }
  
    if((sw = dynamic_cast<Switch*>(n))) next_sync[sw->st] = next_sync[n];

    fk=dynamic_cast<Fork*>(n);
    if(fk) {sync_stack.pop(); }
    
    sy=dynamic_cast<Sync*>(n);
    if(sy){ sync_stack.push(sy); }

  }


  void GRC2zaphod::dfs_signals(GRCNode *n, int step, sigview* cview)
  {

    DefineSignal *defs;
    SignalSymbol *ss;
    zaphod_signal *zs;
    sigview *newview;

    if(!n) return;
    
    if (dynamic_cast<STNode *>(n)) return;
    if(dfs_black.count(n)>0) return;


    if((defs = dynamic_cast<DefineSignal*>(n))){
      if(step == 0) { // build depth signals

	for(SymbolTable::iterator isym = defs->signals->begin();
	    isym != defs->signals->end(); isym++){
	  assert((ss = dynamic_cast<SignalSymbol*>(*isym)));
	  
	  if(flat_local_signals.count(ss) == 0){
	    zs = new zaphod_signal();
	    zs->signal = new wire();
	    zs->realname = ss->name;
	    flat_local_signals[ss]=zs;
	    depth_local_signals[ss]=zs;
	    std::cerr<<"new local signal "<<zs->realname<<" defined for depth\n";
	  }	
	}
      }
    
      if(step == 1) { // build the surface signals
	std::cerr<<"creating new view\n";
	newview = new sigview(*cview); // copy the oldview
	for(SymbolTable::iterator isym = defs->signals->begin();
	    isym != defs->signals->end(); isym++){
	  assert((ss = dynamic_cast<SignalSymbol*>(*isym)));
	  zs = new zaphod_signal(); // create new wire for the signal instance
	  zs->signal = new wire();
	  (*newview)[ss]=zs;
	  ss = new SignalSymbol(*ss);
	  stringstream sstream;
	  sstream << ss->name << refmap[n];
	  ss -> name = sstream.str();
	  zs->realname = ss->name;
	  flat_local_signals[ss] = zs;
	  std::cerr<<"new local signal for "<<zs->realname<<" defined\n";
	}	
	cview = newview;
      }
    }

    localsignals[n] = cview;

    dfs_black.insert(n); // color black
    for(vector<GRCNode*>::iterator i=n->successors.begin(); i!=n->successors.end();i++)
      dfs_signals(*i, step, cview);

  }

  void GRC2zaphod::link_syncs()
  {
    Sync *sy, *nxsy;
    wire *w, *wn;
    int i, tsz;

    tsz=topo.size();
    for(i=0; i<tsz; i++){
      sy=dynamic_cast<Sync*>(topo[i]);
      if(!sy) continue;
      nxsy = next_sync[sy];

      if(sdebug)std::cerr<<"*** SYnc "<<refmap[sy]<<'\n';
     
      w = current ->build_orgate(local_enters_valid[sy], local_enters_valid[nxsy]);
      local_enters_valid[sy] = w;
      if(sdebug)std::cerr<<"    w  localenters = "<<current->wn(w)<<'\n';

      w = current ->build_orgate(w, fin_loop_valid[sy]);
      if(sdebug)std::cerr<<"    w  finloop = "<<current->wn(w)<<'\n';
      fin_loop_valid[sy] = w;

    }

     if(sdebug)std::cerr<<"                BREAK\n";
    
    
    for(i=0; i<tsz; i++){
      sy=dynamic_cast<Sync*>(topo[i]);
      if(!sy) continue;
    
      if(sdebug)std::cerr<<"*** SYnc "<<refmap[sy]<<'\n';

      w = local_enters_valid[sy];
      if(sdebug)std::cerr<<"    w  localenters = "<<current->wn(w)<<'\n';
      wn = current ->build_notgate(w);
      if(sdebug)std::cerr<<"    wn localenters = "<<current->wn(wn)<<'\n';
      local_enters_valid[sy] = wn;
      new strap(wn, "UU", refmap[sy],3);

      w = fin_loop_valid[sy];
      if(sdebug)std::cerr<<"    w  finloop = "<<current->wn(w)<<'\n';
      wn = current ->build_notgate(w);
      if(sdebug)std::cerr<<"    wn finloop = "<<current->wn(wn)<<'\n';
      fin_loop_valid[sy] = wn;
      new strap(wn, "TT", refmap[sy], 3);

    }

    local_enters_valid[NULL] = w_1;
    fin_loop_valid[NULL] = w_1;

  }

  Status GRC2zaphod::visit(Literal &s)
  {
    int x;
    if(edebug)cerr<<"Literal: "<<s.value<<"\n";
    if(sscanf(s.value.c_str(),"%d",&x)!=1) assert(0);
    zc.setint(x,1); // is constant
    return Status();
  }

  
  Status GRC2zaphod::visit(LoadVariableExpression &s)
  {
    cerr<<"load var expr: not supported yet\n";
    assert(0);
    //    s.variable->welcome(*this);
    return Status();
  }

  Status GRC2zaphod::visit(VariableSymbol &s)
  {

    std::cerr<<"var sym's name: "<<s.name<<" why here?\n";
    assert(0);
    //    s.initializer->welcome(*this);
    return Status();
  }

  Status GRC2zaphod::visit(Assignment &s)
  {
    std::cerr<<"Assignment: not supported yet\n";
    assert(0);
    s.value -> welcome(*this);
    return Status();
  }

 Status GRC2zaphod::visit(Emit &s)
  {
    SignalSymbol *ss;
    wire *flow;
    zaphod_signal *sig;

    flow = zc.getflow();
    ss=s.signal;
    //    std::cerr<<"signal"<<ss->name<<" is of type "<<ss->direction<<"\n";
    if(ss->direction == "local") {
      assert(zc.local_sig);
      sig = (*zc.local_sig)[ss];
      if(edebug)std::cerr<<"For local sig "<<ss->name<<" using "<<sig->realname<<"\n";
    } else sig = &signals[ss]; // not nice
    assert(sig);
   
    sig -> sigemit.push_back(flow);

    return Status();
  }

  Status GRC2zaphod::visit(LoadSignalExpression &s)
  {
    SignalSymbol *ss;
    zaphod_signal *sig;

    ss = s.signal;
    //    std::cerr<<"signal"<<ss->name<<" is of type "<<ss->direction<<"\n";
    if(ss->direction == "local") {
      assert(zc.local_sig);
      sig = (*zc.local_sig)[ss];
      if(edebug)std::cerr<<"For local sig "<<ss->name<<" using "<<sig->realname<<"\n";
    } else sig = &signals[ss]; // not nice
    assert(sig);

    if(ss->name == "tick") zc.setbool(w_1); else
      zc.setbool(sig -> signal);
    return Status();
  }

  Status GRC2zaphod::visit(BinaryOp &s)
  {
    if(edebug)cerr<<"binary "<<s.op<<"\n";    

    wire *w1, *w2;
    w1 = zc.getbool(s.source1);
    w2 = zc.getbool(s.source2);

    if(s.op == "and"){
      zc.setbool ( current-> build_andgate(w1, w2));
      return Status();
    }

    if (s.op == "or"){
      zc.setbool ( current->build_orgate(w1,w2));
      return Status();
    }

    std::cerr<<"BinaryOp "<<s.op<<"not supported\n";
    assert(0);

    return Status();
  }
  
  Status GRC2zaphod::visit(UnaryOp &s)
  {
    wire *win;

    win = zc.getbool(s.source);
    if(s.op == "not"){
      zc.setbool(current-> build_notgate(win));
      return Status();
    }

    std::cerr<<"UnaryOp "<<s.op<<"not supported\n";
    assert(0);
    return Status();
  }

  wire* zaphod_counter::get_check()
  {
    wire *w;
    if(check) w = check; else { w = check = new wire(); }
    return w;
  }

  Status GRC2zaphod::visit(StartCounter &s)
  {
    int stval;
    wire *wpred;
    zaphod_counter *cnt;
    STNode *par;

    if(edebug) std::cerr<<"StartCounter: "<<current->wn(zc.getflow())<<"\n";
    par = (STNode*) s.st->predecessors[0];
    if(edebug)std::cerr<<"   ... st is "<<refmap[s.st]<<" will use suspend "<<
		refmap[par]<<"\n";

    stval = zc.getintconst(s.startvalue);
    wpred = zc.getbool(s.predicate);
    cnt = &counters[s.counter];

    cnt->start.push_back(zc.getflow());
    cnt->startval = stval;
    cnt->enable = wpred;
    cnt->get_check();

    cnt->st = par;

    return Status();
  }

  Status GRC2zaphod::visit(CheckCounter &s)
  {
    wire *w;

    if(edebug) std::cerr<<"CheckCounter\n";
    // not nice ! we have to setup counters first by using CounterSymbol
    // now assume a counter is checked only in one place (in the depth)
    w = counters[s.counter].get_check(); 
    zc.setbool(w);

    return Status();
  }

   void GRC2zaphod::build_counters()
   {
     zaphod_counter *cnt;
     wire *wst, *wnsus, *wen;

     for(map<VariableSymbol*, zaphod_counter>::iterator iv = counters.begin();
	 iv != counters.end(); iv++){
       if(edebug)std::cerr<<"Builing new counter ...\n";

       cnt = &(*iv).second;

       assert((int) cnt->start.size() > 0);
       assert(cnt->startval > 0);
       wst = current-> build_orgate(cnt->start);

       wnsus = current->build_notgate(suspend[cnt->st]);
       wen = current->build_andgate(wnsus, cnt->enable);

       if(edebug)std::cerr<<"   w/ alarm="<<current->wn(cnt->check)<<
		   " start="<<current->wn(wst)<<" enable="<<current->wn(wen)<<
		   " startval="<<cnt->startval<<"\n";
       new counter(cnt->check, wst, wen, cnt->startval - 1);

     }
   }

  void zaphod_context::eval_expr(Expression *e)
  {
      t = NULL; isconst = 0;
      e -> welcome(*owner);
  }

  void zaphod_context::do_statement(Statement *s, wire *flowin){
    flow = flowin;
    s->welcome(*owner);
  }

}
