#!/bin/sh

# Simple testing framework for running strlxml (the Esterel front-end) and
# checking that the generated XML is correct by printing it out in Esterel
# form and comparing

STRLXML=./strlxml
XMLSTRL=./xmlstrl

globallog=teststrlxml.log
rm -f $globallog
error=0

CheckIdempotent() {
    basename=`echo $1 | sed 's/.*\\///
                             s/.strl//'`
    reffile=`echo $1 | sed 's/.strl$/.out/'`
    xmlfile=${basename}.xml
    outfile=${basename}.out
    difffile=${basename}.diff
    echo -n "Parsing $basename..."
    echo "Parsing $basename" 1>&2
    $STRLXML < $1 > $xmlfile 2>&1 || {
	echo "FAILED: strlxml terminated"
        error=1
	return 1
    }
    $XMLSTRL < $xmlfile > $outfile 2>&1 || {
	echo "FAILED: xmlstrl terminated"
        error=1
	return 1
    }
    if test -e $reffile; then
	# Compare with reference file
	diff -b $reffile $outfile > $difffile 2>&1 || {
	    echo "FAILED: output mismatch"
            error=1
	    return 1
	}
    else
	# Compare with source
	diff -b $1 $outfile > $difffile 2>&1 || {
	    echo "FAILED: output mismatch"
            error=1
	    return 1
	}
    fi
    rm $xmlfile $outfile $difffile
    echo OK
}

CheckError() {
    basename=`echo $1 | sed 's/.*\\///
                             s/.strl//'`
    logfile=${basename}.log
    echo -n "Parsing $basename..."
    echo "Parsing $basename" 1>&2
    $STRLXML < $1 > $logfile && {
       echo "FAILED: strlxml gave no error"
       error=1
       return 1
    }
    echo OK
}

for file in tests/test*.strl
do
  CheckIdempotent $file 2>> $globallog
done

for file in tests/fail*.strl
do
  CheckError $file 2>> $globallog
done

exit $error
